package cmd

import (
	"context"
	"os"
	"syscall"
	"time"

	"github.com/Masterminds/squirrel"
	"github.com/dlmiddlecote/sqlstats"
	"github.com/go-openapi/loads"
	"github.com/jmoiron/sqlx"
	flags "github.com/orus-io/go-flags"
	"github.com/prometheus/client_golang/prometheus"
	"orus.io/orus-io/go-orusapi/database"

	"orus.io/orus-io/rednerd/auth"
	"orus.io/orus-io/rednerd/migration"
	"orus.io/orus-io/rednerd/models"
	"orus.io/orus-io/rednerd/rendering"
	"orus.io/orus-io/rednerd/restapi"
	"orus.io/orus-io/rednerd/restapi/operations"
	"orus.io/orus-io/rednerd/utils"
)

// NewServerCmd creates a ServerCmd.
func NewServerCmd(options *Options) *ServeCmd {
	swaggerSpec, err := loads.Embedded(restapi.SwaggerJSON, restapi.FlatSwaggerJSON)
	if err != nil {
		log.Fatal().Err(err).Msg("")
	}

	api := operations.NewRednerdAPI(swaggerSpec)
	server := restapi.NewServer(api)

	cmd := ServeCmd{
		Server:       server,
		API:          api,
		options:      options,
		TokenOptions: auth.NewTokenOptions(),
	}
	cmd.Server.ConfigureFlags()

	return &cmd
}

// ServeCmd is the 'serve' command.
type ServeCmd struct {
	Server *restapi.Server

	AutoMigrate bool `long:"auto-migrate" description:"automatically apply database migrations if needed"`

	API     *operations.RednerdAPI `no-flag:"t"`
	options *Options

	TokenOptions *auth.TokenOptions `no-flag:"t"`

	DisableMailSender bool `long:"disable-mail-sender" description:"disable the mail sender routine"`
}

// Execute setup a server and runs it.
func (cmd *ServeCmd) Execute([]string) error {
	ctx, cancel := utils.CtxWithSignal(context.Background(), os.Interrupt, syscall.SIGQUIT, syscall.SIGTERM)
	defer cancel()

	ctx = log.WithContext(ctx)

	if cmd.AutoMigrate {
		if err := database.AutoMigrate(cmd.options.Dsn, migration.Source, log); err != nil {
			return err
		}
	} else {
		if err := database.IsUptodate(cmd.options.Dsn, migration.Source); err != nil {
			return err
		}
	}

	db, err := sqlx.Open("postgres", cmd.options.Dsn)
	if err != nil {
		return err
	}

	if cmd.options.DBMaxConn != 0 {
		db.SetMaxOpenConns(cmd.options.DBMaxConn)
	}

	defer func() {
		if err := db.Close(); err != nil {
			log.Err(err).Msg("could not close database connection properly")
		}
	}()

	if err := db.Ping(); err != nil {
		return err
	}

	if cmd.AutoMigrate {
		var etagMissing bool
		if err := db.GetContext(
			ctx,
			&etagMissing,
			"SELECT EXISTS(SELECT 1 FROM template WHERE etag = '' LIMIT 1)",
		); err != nil {
			return err
		}

		if etagMissing {
			sqlH := database.NewSQLHelper(ctx, db, log)

			var templates []*models.TemplateDB

			if err := sqlH.Select(
				&templates,
				squirrel.Select(models.TemplateDBColumns...).
					From(models.TemplateDBTable).
					Where(squirrel.Eq{models.TemplateDBETagColumn: ""}),
			); err != nil {
				return err
			}

			for _, t := range templates {
				t.FromDB()

				t.ETag = t.CalcETag()
				if err := sqlH.UpdateColumns(t, models.TemplateDBETagColumn); err != nil {
					return err
				}
			}
		}
	}

	if cmd.Server.Prometheus {
		collector := sqlstats.NewStatsCollector("db", db)
		prometheus.MustRegister(collector)
	}

	if err := engineOptions.SetupEngines(log); err != nil {
		return err
	}

	defer rendering.DefaultEngineRegistry.Close()

	cmd.Server.APIConfigurator = restapi.APIConfig{
		Context:                         ctx,
		Log:                             log,
		DB:                              db,
		Now:                             time.Now,
		RenderingEngineRegistry:         rendering.DefaultEngineRegistry,
		TemplateRenderingEngineRegistry: rendering.DefaultTemplateEngineRegistry,
		BaseURL:                         cmd.options.BaseURL,
		TokenOptions:                    cmd.TokenOptions,
	}.ConfigureAPI
	cmd.Server.SetLog(log)
	cmd.Server.ConfigureAPI()

	if !cmd.DisableMailSender {
		mailSenderContext, stopMailSender := context.WithCancel(context.Background())
		mailSenderContext = log.WithContext(mailSenderContext)

		if err := MailSender.Start(mailSenderContext, cmd.options.Dsn, db); err != nil {
			stopMailSender()

			return err
		}

		defer func() {
			stopMailSender()
		}()
	}

	defer func() {
		if err := cmd.Server.Shutdown(); err != nil {
			log.Err(err).Msg("error shutting down the server")
		}
	}()

	if err := cmd.Server.Serve(); err != nil {
		return err
	}

	return nil
}

// AddToParser adds the command to the given parser.
func (cmd *ServeCmd) AddToParser(parser *flags.Parser) error {
	serve, err := parser.AddCommand("serve", "Start the HTTP server", "", cmd)
	if err != nil {
		return err
	}

	serve.EnvNamespace = "SERVE"
	serve.FindOptionByLongName("scheme").EnvDefaultKey = "SCHEME"

	for _, optsGroup := range cmd.API.CommandLineOptionsGroups {
		_, err := serve.AddGroup(optsGroup.ShortDescription, optsGroup.LongDescription, optsGroup.Options)
		if err != nil {
			log.Fatal().Msg(err.Error())
		}
	}

	if _, err := parser.AddGroup("token", "Token options", cmd.TokenOptions); err != nil {
		return err
	}

	return nil
}
